# ruff: noqa: N806, N999, SIM115

import os
import tempfile

from .... import examples as pysal_examples
from ..pyShpIO import PurePyShpWrapper


class TesttestPurePyShpWrapper:
    def setup_method(self):
        test_file = pysal_examples.get_path("10740.shp")
        self.test_file = test_file
        self.shp_obj = PurePyShpWrapper(test_file, "r")
        f = tempfile.NamedTemporaryFile(suffix=".shp")
        shpcopy = f.name
        f.close()
        self.shpcopy = shpcopy
        self.shxcopy = shpcopy.replace(".shp", ".shx")

    def test_len(self):
        assert len(self.shp_obj) == 195

    def test_tell(self):
        assert self.shp_obj.tell() == 0
        self.shp_obj.read(1)
        assert self.shp_obj.tell() == 1
        self.shp_obj.read(50)
        assert self.shp_obj.tell() == 51
        self.shp_obj.read()
        assert self.shp_obj.tell() == 195

    def test_seek(self):
        self.shp_obj.seek(0)
        assert self.shp_obj.tell() == 0
        self.shp_obj.seek(55)
        assert self.shp_obj.tell() == 55
        self.shp_obj.read(1)
        assert self.shp_obj.tell() == 56

    def test_read(self):
        self.shp_obj.seek(0)
        objs = self.shp_obj.read()
        assert len(objs) == 195

        self.shp_obj.seek(0)
        objs_b = list(self.shp_obj)
        assert len(objs_b) == 195

        for shp_a, shp_b in zip(objs, objs_b, strict=True):
            assert shp_a.vertices == shp_b.vertices

    def test_random_access(self):
        self.shp_obj.seek(57)
        shp57 = self.shp_obj.read(1)[0]
        self.shp_obj.seek(32)
        shp32 = self.shp_obj.read(1)[0]

        self.shp_obj.seek(57)
        assert self.shp_obj.read(1)[0].vertices == shp57.vertices
        self.shp_obj.seek(32)
        assert self.shp_obj.read(1)[0].vertices == shp32.vertices

    def test_write(self):
        out = PurePyShpWrapper(self.shpcopy, "w")
        self.shp_obj.seek(0)
        for shp in self.shp_obj:
            out.write(shp)
        out.close()

        orig = open(self.test_file, "rb")
        copy = open(self.shpcopy, "rb")
        assert orig.read() == copy.read()
        orig.close()
        copy.close()

        oshx = open(self.test_file.replace(".shp", ".shx"), "rb")
        cshx = open(self.shxcopy, "rb")
        assert oshx.read() == cshx.read()
        oshx.close()
        cshx.close()

        os.remove(self.shpcopy)
        os.remove(self.shxcopy)



import numpy
import pandas
from scipy import optimize, sparse, spatial, stats

from ._utils import (
    CoplanarError,
    _build_coplanarity_lookup,
    _induce_cliques,
    _jitter_geoms,
    _reorder_adjtable_by_ids,
    _sparse_to_arrays,
    _validate_geometry_input,
)

try:
    from sklearn import metrics, neighbors

    HAS_SKLEARN = True
except ImportError:
    HAS_SKLEARN = False

_VALID_GEOMETRY_TYPES = ["Point"]


def _triangular(distances, bandwidth):
    u = numpy.clip(distances / bandwidth, 0, 1)
    return 1 - u


def _parabolic(distances, bandwidth):
    u = numpy.clip(distances / bandwidth, 0, 1)
    return 0.75 * (1 - u**2)


def _gaussian(distances, bandwidth):
    u = distances / bandwidth
    return numpy.exp(-((u / 2) ** 2)) / (numpy.sqrt(2) * numpy.pi)


def _bisquare(distances, bandwidth):
    u = numpy.clip(distances / bandwidth, 0, 1)
    return (15 / 16) * (1 - u**2) ** 2


def _cosine(distances, bandwidth):
    u = numpy.clip(distances / bandwidth, 0, 1)
    return (numpy.pi / 4) * numpy.cos(numpy.pi / 2 * u)


def _boxcar(distances, bandwidth):
    r = (distances < bandwidth).astype(int)
    return r


def _identity(distances, _):
    return distances


_kernel_functions = {
    "triangular": _triangular,
    "parabolic": _parabolic,
    "gaussian": _gaussian,
    "bisquare": _bisquare,
    "cosine": _cosine,
    "boxcar": _boxcar,
    "discrete": _boxcar,
    "identity": _identity,
    None: _identity,
}


def _kernel(
    coordinates,
    bandwidth=None,
    metric="euclidean",
    kernel="gaussian",
    k=None,
    ids=None,
    p=2,
    taper=True,
    coplanar="raise",
    resolve_isolates=True,
):
    """
    Compute a kernel function over a distance matrix.

    Paramters
    ---------
    coordinates : numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
        geometries over which to compute a kernel. If a geopandas.Geo* object
        is provided, the .geometry attribute is used. If a numpy.ndarray with
        a geometry dtype is used, then the coordinates are extracted and used.
    bandwidth : float (default: None)
        distance to use in the kernel computation. Should be on the same scale as
        the input coordinates.
    metric : string or callable (default: 'euclidean')
        distance function to apply over the input coordinates. Supported options
        depend on whether or not scikit-learn is installed. If so, then any
        distance function supported by scikit-learn is supported here. Otherwise,
        only euclidean, minkowski, and manhattan/cityblock distances are admitted.
    kernel : string or callable (default: 'gaussian')
        kernel function to apply over the distance matrix computed by `metric`.
        The following kernels are supported:
            - triangular:
            - parabolic:
            - gaussian:
            - bisquare:
            - cosine:
            - boxcar/discrete: all distances less than `bandwidth` are 1, and all
                other distances are 0
            - identity/None : do nothing, weight similarity based on raw distance
            - callable : a user-defined function that takes the distance vector and
                the bandwidth and returns the kernel: kernel(distances, bandwidth)
    k : int (default: None)
        number of nearest neighbors used to truncate the kernel. This is assumed
        to be constant across samples. If None, no truncation is conduted.
    ids : numpy.narray (default: None)
        ids to use for each sample in coordinates. Generally, construction functions
        that are accessed via Graph.build_kernel() will set this automatically from
        the index of the input. Do not use this argument directly unless you intend
        to set the indices separately from your input data. Otherwise, use
        data.set_index(ids) to ensure ordering is respected. If None, then the index
        from the input coordinates will be used.
    p : int (default: 2)
        parameter for minkowski metric, ignored if metric != "minkowski".
    taper : bool (default: True)
        remove links with a weight equal to zero
    resolve_isolates : bool
        Try to resolve isolates. Can be disabled if we are dealing with cliques later.
    """
    if metric != "precomputed":
        coordinates, ids, _ = _validate_geometry_input(
            coordinates, ids=ids, valid_geometry_types=_VALID_GEOMETRY_TYPES
        )
    else:
        assert (
            coordinates.shape[0] == coordinates.shape[1]
        ), "coordinates should represent a distance matrix if metric='precomputed'"
        if ids is None:
            ids = numpy.arange(coordinates.shape[0])

    if (
        metric == "haversine"
        and not (
            (coordinates[:, 0] > -180)
            & (coordinates[:, 0] < 180)
            & (coordinates[:, 1] > -90)
            & (coordinates[:, 1] < 90)
        ).all()
    ):
        raise ValueError(
            "'haversine' metric is limited to the range of latitude coordinates "
            "(-90, 90) and the range of longitude coordinates (-180, 180)."
        )

    if k is not None:
        if metric != "precomputed":
            d = _knn(coordinates, k=k, metric=metric, p=p, coplanar=coplanar)
        else:
            d = coordinates * (coordinates.argsort(axis=1, kind="stable") < (k + 1))
    else:
        if metric != "precomputed":
            dist_kwds = {}
            if metric == "minkowski":
                dist_kwds["p"] = p
            if HAS_SKLEARN:
                sq = metrics.pairwise_distances(
                    coordinates, coordinates, metric=metric, **dist_kwds
                )
            else:
                if metric not in ("euclidean", "manhattan", "cityblock", "minkowski"):
                    raise ValueError(
                        f"metric {metric} is not supported by scipy, and scikit-learn "
                        "could not be imported."
                    )
                d = spatial.distance.pdist(coordinates, metric=metric, **dist_kwds)
                sq = spatial.distance.squareform(d)

            # ensure that self-distance is dropped but 0 between co-located pts not
            # get data and ids for sparse constructor
            data = sq.flatten()
            i = numpy.tile(numpy.arange(sq.shape[0]), sq.shape[0])
            j = numpy.repeat(numpy.arange(sq.shape[0]), sq.shape[0])
            # remove diagonal
            data = numpy.delete(data, numpy.arange(0, data.size, sq.shape[0] + 1))
            i = numpy.delete(i, numpy.arange(0, i.size, sq.shape[0] + 1))
            j = numpy.delete(j, numpy.arange(0, j.size, sq.shape[0] + 1))
            # construct sparse
            d = sparse.csc_array((data, (i, j)))
        else:
            d = sparse.csc_array(coordinates)
    if bandwidth is None:
        bandwidth = numpy.percentile(d.data, 25)
    elif bandwidth == "auto":
        if (kernel == "identity") or (kernel is None):
            bandwidth = numpy.nan  # ignored by identity
        else:
            bandwidth = _optimize_bandwidth(d, kernel)
    if callable(kernel):
        d.data = kernel(d.data, bandwidth)
    else:
        d.data = _kernel_functions[kernel](d.data, bandwidth)

    if taper:
        d.eliminate_zeros()

    return _sparse_to_arrays(d, ids=ids, resolve_isolates=resolve_isolates)


def _knn(coordinates, metric="euclidean", k=1, p=2, coplanar="raise"):
    """internal function called only within _kernel, never directly to build KNN"""
    coordinates, ids, geoms = _validate_geometry_input(
        coordinates, ids=None, valid_geometry_types=_VALID_GEOMETRY_TYPES
    )
    if coplanar == "jitter":
        coordinates, geoms = _jitter_geoms(coordinates, geoms=geoms)

    n_coplanar = geoms.geometry.duplicated().sum()
    n_samples, _ = coordinates.shape

    if n_coplanar == 0:
        if metric == "haversine":
            # sklearn haversine works with (lat,lng) in radians...
            coordinates = numpy.fliplr(numpy.deg2rad(coordinates))
        query = _prepare_tree_query(coordinates, metric, p=p)
        d_linear, ixs = query(coordinates, k=k + 1)
        self_ix, neighbor_ix = ixs[:, 0], ixs[:, 1:]
        d_linear = d_linear[:, 1:]
        self_ix_flat = numpy.repeat(self_ix, k)
        neighbor_ix_flat = neighbor_ix.flatten()
        d_linear_flat = d_linear.flatten()
        if metric == "haversine":
            d_linear_flat * 6371  # express haversine distances in kilometers
        d = sparse.csr_array(
            (d_linear_flat, (self_ix_flat, neighbor_ix_flat)),
            shape=(n_samples, n_samples),
        )
        return d
    else:
        coplanar_lookup, nearest = _build_coplanarity_lookup(geoms)
        _, counts = numpy.unique(nearest, return_counts=True)
        max_at_one_site = counts.max()
        if coplanar == "raise":
            raise CoplanarError(
                f"There are {len(coordinates) - len(coplanar_lookup)} unique locations "
                f"in the dataset, but {len(coordinates)} observations. At least one of "
                f"these sites has {max_at_one_site} points, more than the {k} nearest "
                f"neighbors requested. This means there are more than {k} points in "
                "the same location, which makes this graph type undefined. To address "
                "this issue, consider setting `coplanar='clique'` or consult the "
                "documentation about coplanar points."
            )
        if coplanar == "jitter":
            # force re-jittering over and over again until the coplanarity is broken
            return _knn(
                _jitter_geoms(coordinates, geoms)[-1],
                metric=metric,
                k=k,
                p=p,
                coplanar="jitter",
            )

        if coplanar == "clique":
            heads, tails, weights = _sparse_to_arrays(
                _knn(
                    numpy.delete(coordinates, coplanar_lookup, 0),
                    metric=metric,
                    k=k,
                    p=p,
                    coplanar="raise",
                )
            )
            adjtable = pandas.DataFrame.from_dict(
                {"focal": heads, "neighbor": tails, "weight": weights}
            )
            adjtable = _induce_cliques(
                adjtable, coplanar_lookup, nearest, fill_value=-1
            )
            adjtable["focal"] = ids[adjtable.focal]
            adjtable["neighbor"] = ids[adjtable.neighbor]
            adjtable = _reorder_adjtable_by_ids(adjtable, ids)
            sparse_out = sparse.csr_array(
                (
                    adjtable.weight.values,
                    (adjtable.focal.values, adjtable.neighbor.values),
                ),
                shape=(n_samples, n_samples),
            )
            sparse_out.data[sparse_out.data < 0] = 0
            return sparse_out
        raise ValueError(
            f"'{coplanar}' is not a valid option. Use one of "
            "['raise', 'jitter', 'clique']."
        )


def _distance_band(coordinates, threshold, ids=None):
    coordinates, ids, _ = _validate_geometry_input(
        coordinates, ids=ids, valid_geometry_types=_VALID_GEOMETRY_TYPES
    )
    tree = spatial.KDTree(coordinates)
    sp = sparse.csr_array(tree.sparse_distance_matrix(tree, threshold))
    return sp


def _prepare_tree_query(coordinates, metric, p=2):
    """
    Construct a tree query function relevant to the input metric.
    Prefer scikit-learn trees if they are available.
    """
    if HAS_SKLEARN:
        dist_kwds = {}
        if metric == "minkowski":
            dist_kwds["p"] = p
        if metric in neighbors.VALID_METRICS["kd_tree"]:
            tree = neighbors.KDTree
        else:
            tree = neighbors.BallTree
        return tree(coordinates, metric=metric, **dist_kwds).query
    else:
        if metric in ("euclidean", "manhattan", "cityblock", "minkowski"):
            tree_ = spatial.KDTree(coordinates)
            p = {"euclidean": 2, "manhattan": 1, "cityblock": 1, "minkowski": p}[metric]

            def query(target, k):
                return tree_.query(target, k=k, p=p)

            return query
        else:
            raise ValueError(
                f"metric {metric} is not supported by scipy, and scikit-learn could "
                "not be imported"
            )


def _optimize_bandwidth(d, kernel):
    """
    Optimize the bandwidth as a function of entropy for a given kernel function.

    This ensures that the entropy of the kernel is maximized for a given
    distance matrix. This will result in the smoothing that provide the most
    uniform distribution of kernel values, which is a good proxy for a
    "moderate" level of smoothing.
    """
    kernel_function = _kernel_functions.get(kernel, kernel)
    assert callable(kernel_function), (
        f"kernel {kernel} was not in supported kernel types "
        f"{_kernel_functions.keys()} or callable"
    )

    def _loss(bandwidth, d=d, kernel_function=kernel_function):
        k_u = kernel_function(d.data, bandwidth)
        bins, _ = numpy.histogram(k_u, bins=int(d.shape[0] ** 0.5), range=(0, 1))
        return -stats.entropy(bins / bins.sum())

    xopt = optimize.minimize_scalar(
        _loss, bounds=(0, d.data.max() * 2), method="bounded"
    )
    return xopt.x



"""Handle remote datasets."""

import warnings

import requests
from bs4 import BeautifulSoup

from .base import Example

# remote_dict holds the metadata for remote datasets from the geoda center
# to update prior to release run _remote_data()

_remote_dict = {
    "AirBnB": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/airbnb.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//airbnb/",
        "n": "77",
        "k": "20",
        "description": "Airbnb rentals, socioeconomics, and crime in Chicago",
    },
    "Atlanta": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab/" "/data/atlanta_hom.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//atlanta_old/",
        "n": "90",
        "k": "23",
        "description": "Atlanta, GA region homicide counts and rates",
    },
    "Baltimore": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/baltimore.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//baltim/",
        "n": "211",
        "k": "17",
        "description": "Baltimore house sales prices and hedonics",
    },
    "Bostonhsg": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/boston.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//boston-housing/",
        "n": "506",
        "k": "23",
        "description": "Boston housing and neighborhood data",
    },
    "Buenosaires": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/buenosaires.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//buenos-aires_old/",
        "n": " 209",
        "k": " 21",
        "description": " Electoral Data for 1999 Argentinean Elections",
    },
    "Cars": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab/"
            "/data/Abandoned_Vehicles_Map.csv"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab//1-source-and-description/"
        ),
        "n": "137,867",
        "k": "21",
        "description": "2011 abandoned vehicles in Chicago (311 complaints).",
    },
    "Charleston1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/CharlestonMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//charleston-1_old/",
        "n": " 117",
        "k": " 30",
        "description": " 2000 Census Tract Data for Charleston, SC MSA and counties",
    },
    "Charleston2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/CharlestonMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//charleston2/",
        "n": " 44",
        "k": " 97",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns (Census Bureau)"
            " for Charleston, SC MSA"
        ),
    },
    "Chicago Health": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/comarea.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//comarea_vars/",
        "n": " 77",
        "k": " 86",
        "description": " Chicago Health + Socio-Economics",
    },
    "Chicago commpop": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/chicago_commpop.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//commpop/",
        "n": " 77",
        "k": " 8",
        "description": (
            " Chicago Community Area Population Percent Change for 2000 and 2010"
        ),
    },
    "Chicago parcels": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab/"
            "/https://uchicago.box.com/s/j2d2ch5uvckse24y8l7vh9198wnq216i"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//parcels/",
        "n": " 592,521",
        "k": " 5",
        "description": " Tax parcel polygons of Cook county",
    },
    "Chile Labor": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/flma.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//FLMA/",
        "n": "141",
        "k": "62",
        "description": "Labor Markets in Chile (1982-2002)",
    },
    "Chile Migration": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab/"
            "/https://uchicago.box.com/s/yqc97nq23hoeeqo5lkc2grlg98skokgk"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//CHIM/",
        "n": " 304",
        "k": " 10",
        "description": " Internal Migration in Chile (1977-2002)",
    },
    "Cincinnati": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/walnuthills_updated.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//walnut_hills/",
        "n": " 457",
        "k": " 89",
        "description": " 2008 Cincinnati Crime + Socio-Demographics",
    },
    "Cleveland": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/cleveland.zip"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab//clev_sls_154_core/"
        ),
        "n": " 205",
        "k": " 9",
        "description": " 2015 sales prices of homes in Cleveland, OH.",
    },
    "Columbus": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/columbus.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//columbus/",
        "n": " 49",
        "k": " 20",
        "description": " Columbus neighborhood crime",
    },
    "Elections": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/election.zip"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab/"
            "/county_election_2012_2016-variables/"
        ),
        "n": " 3,108",
        "k": " 74",
        "description": " 2012 and 2016 Presidential Elections",
    },
    "Grid100": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/grid100.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//grid100/",
        "n": " 100",
        "k": " 34",
        "description": " Grid with simulated variables",
    },
    "Groceries": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/grocery.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//chicago_sup_vars/",
        "n": " 148",
        "k": " 7",
        "description": " 2015 Chicago supermarkets",
    },
    "Guerry": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/guerry.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//Guerry/",
        "n": " 85",
        "k": " 23",
        "description": " Moral statistics of France (Guerry, 1833)",
    },
    "Health+": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/income_diversity.zip"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab/"
            "/co_income_diversity_variables/"
        ),
        "n": " 3,984",
        "k": " 64",
        "description": " 2000 Health, Income + Diversity",
    },
    "Health Indicators": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/healthIndicators.zip"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab//healthindicators-variables/"
        ),
        "n": " 77",
        "k": " 31",
        "description": " Chicago Health Indicators (2005-11)",
    },
    "Hickory1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/HickoryMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//hickory1/",
        "n": " 68",
        "k": " 30",
        "description": " 2000 Census Tract Data for Hickory, NC MSA and counties",
    },
    "Hickory2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/HickoryMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//hickory2/",
        "n": " 29",
        "k": " 55",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns "
            "(Census Bureau) for Hickory, NC MSA"
        ),
    },
    "Home Sales": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/kingcounty.zip"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab//KingCounty-HouseSales2015/"
        ),
        "n": " 21,613",
        "k": " 21",
        "description": " 2014-15 Home Sales in King County, WA",
    },
    "Houston": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/houston_hom.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//houston/",
        "n": " 52",
        "k": " 23",
        "description": " Houston, TX region homicide counts and rates",
    },
    "Juvenile": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/juvenile.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//juvenile/",
        "n": " 168",
        "k": " 3",
        "description": " Cardiff juvenile delinquent residences",
    },
    "Lansing1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/LansingMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//lansing1/",
        "n": " 117",
        "k": " 30",
        "description": " 2000 Census Tract Data for Lansing, MI MSA and counties",
    },
    "Lansing2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/LansingMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//lansing2/",
        "n": " 46",
        "k": " 55",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns "
            "(Census Bureau) for Lansing, MI MSA"
        ),
    },
    "Laozone": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/laozone.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//ozone/",
        "n": " 32",
        "k": " 8",
        "description": " Ozone measures at monitoring stations in Los Angeles basin",
    },
    "LasRosas": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/lasrosas.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//lasrosas/",
        "n": " 1,738",
        "k": " 34",
        "description": (
            " Corn yield, fertilizer and field data for precision "
            "agriculture, Argentina, 1999"
        ),
    },
    "Liquor Stores": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/liquor.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//liq_chicago/",
        "n": " 571",
        "k": " 2",
        "description": " 2015 Chicago Liquor Stores",
    },
    "Malaria": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/malariacolomb.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//colomb_malaria/",
        "n": " 1,068",
        "k": " 50",
        "description": (
            " Malaria incidence and population (1973, 95, 93 "
            "censuses and projections until 2005) \xa0 \xa0 \xa0"
        ),
    },
    "Milwaukee1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/MilwaukeeMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//milwaukee1/",
        "n": " 417",
        "k": " 31",
        "description": " 2000 Census Tract Data for Milwaukee, WI MSA",
    },
    "Milwaukee2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/MilwaukeeMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//milwaukee2/",
        "n": " 83",
        "k": " 55",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns "
            "(Census Bureau) for Milwaukee, WI MSA"
        ),
    },
    "NCOVR": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/ncovr.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//ncovr/",
        "n": "3,085",
        "k": " 69",
        "description": " US county homicides 1960-1990",
    },
    "Natregimes": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/natregimes.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//natregimes/",
        "n": " 3,085",
        "k": " 73",
        "description": " NCOVR with regimes (book/PySAL)",
    },
    "NDVI": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/ndvi.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//ndvi/",
        "n": " 49",
        "k": " 5",
        "description": " Normalized Difference Vegetation Index grid",
    },
    "Nepal": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/nepal.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//nepal/",
        "n": " 75",
        "k": " 61",
        "description": " Health, poverty and education indicators for Nepal districts",
    },
    "NYC": {
        "download_url": "https://geodacenter.github.io/data-and-lab///data/nyc.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//nyc/",
        "n": " 55",
        "k": " 34",
        "description": (
            " Demographic and housing data for New York City subboroughs, 2002-09"
        ),
    },
    "NYC Earnings": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/lehd.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//LEHD_Data/",
        "n": " 108,487",
        "k": " 70",
        "description": " Block-level Earnings in NYC (2002-14)",
    },
    "NYC Education": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/nyc_2000Census.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//NYC-Census-2000/",
        "n": " 2,216",
        "k": " 56",
        "description": " NYC Education (2000)",
    },
    "NYC Neighborhoods": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/nycnhood_acs.zip"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab//NYC-Nhood-ACS-2008-12/"
        ),
        "n": " 195",
        "k": " 98",
        "description": " Demographics for New York City neighborhoods",
    },
    "NYC Socio-Demographics": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/nyctract_acs.zip"
        ),
        "explain_url": (
            "https://geodacenter.github.io/data-and-lab//NYC_Tract_ACS2008_12/"
        ),
        "n": " 2,166",
        "k": " 113",
        "description": " NYC Education + Socio-Demographics",
    },
    "Ohiolung": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/ohiolung.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//ohiolung/",
        "n": " 88",
        "k": " 42",
        "description": " Ohio lung cancer data, 1968, 1978, 1988",
    },
    "Orlando1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/OrlandoMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//orlando1/",
        "n": " 328",
        "k": " 30",
        "description": " 2000 Census Tract Data for Orlando, FL MSA and counties",
    },
    "Orlando2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/OrlandoMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//orlando2/",
        "n": " 94",
        "k": " 59",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns "
            "(Census Bureau) for Orlando, FL MSA"
        ),
    },
    "Oz9799": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/oz9799.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//oz96/",
        "n": " 30",
        "k": " 78",
        "description": " Monthly ozone data, 1997-99",
    },
    "Phoenix ACS": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/phx2.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//phx/",
        "n": " 685",
        "k": " 17",
        "description": (
            " Phoenix American Community Survey Data (2010, 5-year averages)"
        ),
    },
    "Pittsburgh": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/pittsburgh.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//pitt93/",
        "n": " 143",
        "k": " 8",
        "description": " Pittsburgh homicide locations",
    },
    "Police": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/police.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//police/",
        "n": " 82",
        "k": " 21",
        "description": " Police expenditures Mississippi counties",
    },
    "Sacramento1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/sacramento.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//sacramento1/",
        "n": " 403",
        "k": " 30",
        "description": " 2000 Census Tract Data for Sacramento MSA",
    },
    "Sacramento2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/SacramentoMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//sacramento2/",
        "n": " 125",
        "k": " 53",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns "
            "(Census Bureau) for Sacramento MSA"
        ),
    },
    "SanFran Crime": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab/"
            "/data/SFCrime_July_Dec2012.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//SFcrimes_vars/",
        "n": " 3,384",
        "k": " 13",
        "description": (
            " July-Dec 2012 crime incidents in San Francisco "
            "(points + area) - for CAST"
        ),
    },
    "Savannah1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/SavannahMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//savannah1/",
        "n": " 77",
        "k": " 30",
        "description": " 2000 Census Tract Data for Savannah, GA MSA and counties",
    },
    "Savannah2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/SavannahMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//savannah2/",
        "n": " 24",
        "k": " 55",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns "
            "(Census Bureau) for Savannah, GA MSA"
        ),
    },
    "Scotlip": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/scotlip.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//scotlip/",
        "n": " 56",
        "k": " 11",
        "description": " Male lip cancer in Scotland, 1975-80",
    },
    "Seattle1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/SeattleMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//seattle1/",
        "n": " 664",
        "k": " 30",
        "description": " 2000 Census Tract Data for Seattle, WA MSA and counties",
    },
    "Seattle2": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/SeattleMSA2.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//seattle2/",
        "n": " 145",
        "k": " 59",
        "description": (
            " 1998 and 2001 Zip Code Business Patterns "
            "(Census Bureau) for Seattle, WA MSA"
        ),
    },
    "SIDS": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/sids.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//sids/",
        "n": " 100",
        "k": " 13",
        "description": " North Carolina county SIDS death counts",
    },
    "SIDS2": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/sids2.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//sids2/",
        "n": " 100",
        "k": " 17",
        "description": " North Carolina county SIDS death counts and rates",
    },
    "Snow": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/snow.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//snow/",
        "n": " NA",
        "k": " NA",
        "description": " John Snow & the 19th Century Cholera Epidemic",
    },
    "South": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/south.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//south/",
        "n": " 1,412",
        "k": " 69",
        "description": " US Southern county homicides 1960-1990",
    },
    "Spirals": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/spirals.csv",
        "explain_url": "https://geodacenter.github.io/data-and-lab//spirals/",
        "n": " 301",
        "k": " 2",
        "description": " Synthetic spiral points",
    },
    "StLouis": {
        "download_url": "https://geodacenter.github.io/data-and-lab//data/stlouis.zip",
        "explain_url": "https://geodacenter.github.io/data-and-lab//stlouis/",
        "n": " 78",
        "k": " 23",
        "description": " St Louis region county homicide counts and rates",
    },
    "Tampa1": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/TampaMSA.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//tampa1/",
        "n": " 547",
        "k": " 30",
        "description": " 2000 Census Tract Data for Tampa, FL MSA and counties",
    },
    "US SDOH": {
        "download_url": (
            "https://geodacenter.github.io/data-and-lab//data/us-sdoh-2014.zip"
        ),
        "explain_url": "https://geodacenter.github.io/data-and-lab//us-sdoh/",
        "n": " 71,901",
        "k": " 25",
        "description": " 2014 US Social Determinants of Health Data",
    },
}


def _remote_data():
    """Helper function to get remote metadata for each release.

    Returns
    -------
    datasets : dict
        Remote data sets keyed by the dataset name.
        Values are dictionaries with the following keys
        'download_url', 'explain_url', 'n', 'k', 'description'.
    """

    url = "https://geodacenter.github.io/data-and-lab//"
    try:
        page = requests.get(url, timeout=(10, None))
    except:  # noqa: E722
        warnings.warn("Remote data sets not available. Check connection.")  # noqa: B028
        return {}
    soup = BeautifulSoup(page.text, "html.parser")
    samples = soup.find(class_="samples")
    rows = samples.find_all("tr")
    datasets = {}
    for row in rows[1:]:
        data = row.find_all("td")
        name = data[0].text.strip()
        description = data[1].text
        n = data[2].text
        k = data[3].text
        targets = row.find_all("a")
        download_url = url + targets[1].attrs["href"]
        explain_url = url + targets[0].attrs["href"]
        datasets[name] = {
            "download_url": download_url,
            "explain_url": explain_url,
            "n": n,
            "k": k,
            "description": description,
        }
    return datasets


def _build_remotes():
    """Build remote meta data.

    Returns
    -------
    datasets : dict
        Example datasets keyed by the dataset name.
    """

    datasets = {}
    for name in _remote_dict:
        description = _remote_dict[name]["description"]
        n = _remote_dict[name]["n"]
        k = _remote_dict[name]["k"]
        download_url = _remote_dict[name]["download_url"]
        explain_url = _remote_dict[name]["explain_url"]
        datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # Other Remotes
    # rio
    name = "Rio Grande do Sul"
    description = "Cities of the Brazilian State of Rio Grande do Sul"
    n = 497
    k = 3
    download_url = "https://github.com/sjsrey/rio_grande_do_sul/archive/master.zip"
    explain_url = (
        "https://raw.githubusercontent.com/sjsrey/rio_grande_do_sul/master/README.md"
    )
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # nyc bikes
    name = "nyc_bikes"
    description = "New York City Bike Trips"
    n = 14042
    k = 27
    download_url = "https://github.com/sjsrey/nyc_bikes/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/nyc_bikes/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # taz
    name = "taz"
    description = "Traffic Analysis Zones in So. California"
    n = 4109
    k = 14
    download_url = "https://github.com/sjsrey/taz/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/taz/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # clearwater
    name = "clearwater"
    description = "mgwr testing dataset"
    n = 239
    k = 14
    download_url = "https://github.com/sjsrey/clearwater/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/clearwater/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # newHaven
    name = "newHaven"
    description = "Network testing dataset"
    n = 3293
    k = 5
    download_url = "https://github.com/sjsrey/newHaven/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/newHaven/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # Chicago SDOH
    name = "chicagoSDOH"
    description = "Chicago census tract SDOH variables"
    n = 791
    k = 65
    download_url = "https://github.com/lanselin/spreg_sample_data/archive/master.zip"
    explain_url = (
        "https://raw.githubusercontent.com/lanselin/spreg_sample_data/master/README.md"
    )
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # remove Cars dataset as it is broken
    datasets.pop("Cars")

    return datasets


class Remotes:
    """Remote datasets."""

    def __init__(self):
        """Initialize Remotes."""
        self._datasets = None

    @property
    def datasets(self):
        """Create dictionary of remotes."""
        if self._datasets is None:
            self._datasets = _build_remotes()
        return self._datasets


datasets = Remotes()



# ruff: noqa: SIM115

import numpy as np
from scipy import sparse

from ...weights.weights import WSP, W
from .. import fileio

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["GalIO"]


class GalIO(fileio.FileIO):
    """Opens, reads, and writes file objects in `GAL` format."""

    FORMATS = ["gal"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        self._typ = str
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.file = open(self.dataPath, self.mode)

    def read(self, n=-1, sparse=False):  # noqa: ARG002
        """Read in a ``.gal`` file.

        Parameters
        ----------
        n : int
            Read at most ``n`` objects. Default is ``-1``.
        sparse: bool
            If ``True`` return a ``scipy`` sparse object. If ``False``
            return PySAL `W` object. Default is ``False``.

        Returns
        -------
        w : {libpysal.weights.W, libpysal.weights.WSP}
            A PySAL `W` object or a thin PySAL `WSP`.

        """
        self._sparse = sparse
        self._complain_ifclosed(self.closed)

        w = self._read()

        return w

    def seek(self, pos):
        if pos == 0:
            self.file.seek(0)
            self.pos = 0

    def _get_data_type(self):
        return self._typ

    def _set_data_type(self, typ):
        """

        Raises
        ------
        TypeError
            Raised when ``typ`` is not a callable.

        """
        if callable(typ):
            self._typ = typ
        else:
            raise TypeError("Expecting a callable.")

    data_type = property(fset=_set_data_type, fget=_get_data_type)

    def _read(self):
        """Reads in a `GalIO` object.

        Returns
        -------
        w : {libpysal.weights.W, libpysal.weights.WSP}
            A PySAL `W` object or a thin PySAL `WSP`.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Examples
        --------

        >>> import tempfile, libpysal, os

        Read in a file `GAL` file.

        >>> testfile = libpysal.io.open(libpysal.examples.get_path('sids2.gal'), 'r')

        Return a `W` object.

        >>> w = testfile.read()
        >>> w.n == 100
        True

        >>> print(round(w.sd,6))
        1.515124

        >>> testfile = libpysal.io.open(libpysal.examples.get_path('sids2.gal'), 'r')

        Return a sparse matrix for the `W` information.

        >>> wsp = testfile.read(sparse=True)
        >>> wsp.sparse.nnz
        462

        """

        if self._sparse:
            if self.pos > 0:
                raise StopIteration

            header = self.file.readline().strip().split()
            header_n = len(header)
            n = int(header[0])

            if header_n > 1:
                n = int(header[1])

            ids = []
            idsappend = ids.append
            row = []
            extend = row.extend  # avoid dot in loops
            col = []
            append = col.append
            counter = 0
            typ = self.data_type

            for _ in range(n):
                id_, n_neighbors = self.file.readline().strip().split()
                id_ = typ(id)
                n_neighbors = int(n_neighbors)
                neighbors_i = list(map(typ, self.file.readline().strip().split()))
                nn = len(neighbors_i)
                extend([id_] * nn)
                counter += nn

                for id_neigh in neighbors_i:
                    append(id_neigh)
                idsappend(id_)

            self.pos += 1
            row = np.array(row)
            col = np.array(col)
            data = np.ones(counter)
            ids = np.unique(row)
            row = np.array([np.where(ids == j)[0] for j in row]).flatten()
            col = np.array([np.where(ids == j)[0] for j in col]).flatten()
            spmat = sparse.csr_matrix((data, (row, col)), shape=(n, n))

            w = WSP(spmat)

        else:
            if self.pos > 0:
                raise StopIteration

            neighbors = {}
            ids = []

            # handle case where more than n is specified in first line
            header = self.file.readline().strip().split()
            header_n = len(header)
            n = int(header[0])

            if header_n > 1:
                n = int(header[1])

            w = {}
            typ = self.data_type

            for _ in range(n):
                id_, n_neighbors = self.file.readline().strip().split()
                id_ = typ(id_)
                n_neighbors = int(n_neighbors)
                neighbors_i = list(map(typ, self.file.readline().strip().split()))
                neighbors[id_] = neighbors_i
                ids.append(id_)

            self.pos += 1

            w = W(neighbors, id_order=ids)

        return w

    def write(self, obj):
        """Write a weights object to the opened `GAL` file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(libpysal.examples.get_path('sids2.gal'), 'r')
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.gal')

        Reassign to the new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w')

        Write the weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created gal file.

        >>> wnew =  libpysal.io.open(fname, 'r').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            ids = obj.id_order
            self.file.write("%d\n" % (obj.n))

            for id_ in ids:
                neighbors = obj.neighbors[id_]
                self.file.write("%s %d\n" % (str(id_), len(neighbors)))
                self.file.write(" ".join(map(str, neighbors)) + "\n")
            self.pos += 1
        else:
            raise TypeError(f"Expected a PySAL weights object, got: {type(obj)}.")

    def close(self):
        self.file.close()
        fileio.FileIO.close(self)



"""Segment Locator Unittest."""

# ruff: noqa: F403, F405, N999

from ..segmentLocator import *
from ..shapes import *


class TestSegmentGrid:
    def setup_method(self):
        # 10x10 grid with four line segments, one for each edge of the grid.
        self.grid = SegmentGrid(Rectangle(0, 0, 10, 10), 1)
        self.grid.add(LineSegment(Point((0.0, 0.0)), Point((0.0, 10.0))), 0)
        self.grid.add(LineSegment(Point((0.0, 10.0)), Point((10.0, 10.0))), 1)
        self.grid.add(LineSegment(Point((10.0, 10.0)), Point((10.0, 0.0))), 2)
        self.grid.add(LineSegment(Point((10.0, 0.0)), Point((0.0, 0.0))), 3)

    def test_nearest_1(self):
        # Center
        assert self.grid.nearest(Point((5.0, 5.0))) == [0, 1, 2, 3]
        # Left Edge
        assert self.grid.nearest(Point((0.0, 5.0))) == [0]
        # Top Edge
        assert self.grid.nearest(Point((5.0, 10.0))) == [1]
        # Right Edge
        assert self.grid.nearest(Point((10.0, 5.0))) == [2]
        # Bottom Edge
        assert self.grid.nearest(Point((5.0, 0.0))) == [3]

    def test_nearest_2(self):
        # Left Edge
        assert self.grid.nearest(Point((-100000.0, 5.0))) == [0, 1, 3]
        # Right Edge
        assert self.grid.nearest(Point((100000.0, 5.0))) == [1, 2, 3]
        # Bottom Edge
        assert self.grid.nearest(Point((5.0, -100000.0))) == [0, 2, 3]
        # Top Edge
        assert self.grid.nearest(Point((5.0, 100000.0))) == [0, 1, 2]



# ruff: noqa: SIM115

from ...weights import W
from .. import fileio

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["StataTextIO"]


class StataTextIO(fileio.FileIO):
    """Opens, reads, and writes weights file objects in STATA text format.

    Spatial weights objects in the STATA text format are used in STATA ``sppack``
    library through the ``spmat`` command. This format is a simple text file
    delimited by a whitespace. The ``spmat`` command does not specify which file
    extension to use. But, ``.txt`` seems the default file extension, which is
    assumed in PySAL.

    The first line of the STATA text file is a header including the number of
    observations. After this header line, it includes at least one data column
    that contains unique IDs or record numbers of observations. When an ID
    variable is not specified for the original spatial weights matrix in STATA,
    record numbers are used to identify individual observations, and the record
    numbers start with 1. The ``spmat`` command seems to allow only integer IDs,
    which is also assumed in PySAL.

    A STATA text file can have one of the following structures according to
    its export options in STATA.

    Structure 1: Encoding using the list of neighbor IDs.

    ```
    [Line 1]    [Number_of_Observations]
    [Line 2]    [ID_of_Obs_1] [ID_of_Neighbor_1_of_Obs_1] [ID_of_Neighbor_2_of_Obs_1] ... [ID_of_Neighbor_m_of_Obs_1]
    [Line 3]    [ID_of_Obs_2]
    [Line 4]    [ID_of_Obs_3] [ID_of_Neighbor_1_of_Obs_3] [ID_of_Neighbor_2_of_Obs_3]
    ...
    ```

    Note that for island observations their IDs are still recorded.

    Structure 2: Encoding using a full matrix format.

    ```
    [Line 1]    [Number_of_Observations]
    [Line 2]    [ID_of_Obs_1] [w_11] [w_12] ... [w_1n]
    [Line 3]    [ID_of_Obs_2] [w_21] [w_22] ... [w_2n]
    [Line 4]    [ID_of_Obs_3] [w_31] [w_32] ... [w_3n]
    ...
    [Line n+1]  [ID_of_Obs_n] [w_n1] [w_n2] ... [w_nn]
    ```

    where :math:`w_{ij}` can be a form of general weight. That is, :math:`w_ij`
    can be both a binary value or a general numeric value. If an observation
    is an island, all of its ``w`` columns contain 0.

    References
    ----------

    Drukker D.M., Peng H., Prucha I.R., and Raciborski R. (2011)
    "Creating and managing spatial-weighting matrices using the spmat command"

    Notes
    -----

    The ``spmat`` command allows users to add any note to a spatial weights
    matrix object in STATA. However, all those notes are lost when the matrix
    is exported. PySAL also does not take care of those notes.

    """  # noqa: E501

    FORMATS = ["stata_text"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        args = args[:2]
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.file = open(self.dataPath, self.mode)

    def read(self, n=-1):  # noqa: ARG002
        """

        Parameters
        ----------
        n : int
            Read at most ``n`` objects. Default is ``-1``.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        """

        self._complain_ifclosed(self.closed)
        return self._read()

    def seek(self, pos):
        if pos == 0:
            self.file.seek(0)
            self.pos = 0

    def _read(self):
        """Reads STATA Text file
        Returns a pysal.weights.weights.W object

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open a text file and read it into a PySAL weights object.

        >>> import libpysal
        >>> w = libpysal.io.open(
        ...     libpysal.examples.get_path('stata_sparse.txt'), 'r', 'stata_text'
        ... ).read()

        Get the number of observations from the header.

        >>> w.n
        56

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        4.0

        Get neighbor distances for a single observation.

        >>> w[1] == dict({53: 1.0, 51: 1.0, 45: 1.0, 54: 1.0, 7: 1.0})
        True

        """

        if self.pos > 0:
            raise StopIteration

        n = int(self.file.readline().strip())
        line1 = self.file.readline().strip()
        obs_01 = line1.split(" ")
        matrix_form = False

        if len(obs_01) == 1 or float(obs_01[1]) != 0.0:

            def line2wgt(line):
                row = [int(i) for i in line.strip().split(" ")]
                return row[0], row[1:], [1.0] * len(row[1:])

        else:
            matrix_form = True

            def line2wgt(line):
                row = line.strip().split(" ")
                obs = int(float(row[0]))
                ngh, wgt = [], []
                for i in range(n):
                    w = float(row[i + 1])
                    if w > 0:
                        ngh.append(i)
                        wgt.append(w)
                return obs, ngh, wgt

        id_order = []
        weights, neighbors = {}, {}
        l_ = line1

        for _ in range(n):
            obs, ngh, wgt = line2wgt(l_)
            id_order.append(obs)
            neighbors[obs] = ngh
            weights[obs] = wgt
            l_ = self.file.readline()

        if matrix_form:
            for obs in neighbors:
                neighbors[obs] = [id_order[ngh] for ngh in neighbors[obs]]

        self.pos += 1

        w = W(neighbors, weights)

        return w

    def write(self, obj, matrix_form=False):
        """Write a weights object to an opened text file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.
        matrix_form : bool
            Flag for matrix form (``True``). Default is ``False``.

        Raises
        ------
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(
        ...     libpysal.examples.get_path('stata_sparse.txt'), 'r', 'stata_text'
        ... )
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.txt')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w', 'stata_text')

        Write the weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created text file.

        >>> wnew =  libpysal.io.open(fname, 'r', 'stata_text').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            header = f"{obj.n}\n"
            self.file.write(header)
            if matrix_form:

                def wgt2line(obs_id, neighbor, weight):
                    w = ["0.0"] * obj.n
                    for ngh, wgt in zip(neighbor, weight, strict=True):
                        w[obj.id2i[ngh]] = str(wgt)
                    return [str(obs_id)] + w

            else:

                def wgt2line(obs_id, neighbor, _):
                    return [str(obs_id)] + [str(ngh) for ngh in neighbor]

            for id_ in obj.id_order:
                line = wgt2line(id_, obj.neighbors[id_], obj.weights[id_])
                self.file.write("{}\n".format(" ".join(line)))
        else:
            raise TypeError(f"Expected a PySAL weights object, got: {type(obj)}.")

    def close(self):
        self.file.close()
        fileio.FileIO.close(self)



import copy
import importlib

import pandas  #  noqa: F401

try:
    from patsy import PatsyError
except ImportError:
    PatsyError = Exception

RTOL = 0.00001
ATOL = 1e-7
MISSINGVALUE = None

####################
# Decorators/Utils #
####################

# import numba.jit OR create mimic decorator and set existence flag
try:
    from numba import jit

    HAS_JIT = True

except ImportError:

    def jit(function=None, **kwargs):  # noqa: ARG001
        """Mimic numba.jit() with synthetic wrapper."""

        if function is not None:

            def wrapped(*original_args, **original_kw):
                """Case 1 - structure of a standard decorator
                i.e., jit(function)(*args, **kwargs).
                """

                return function(*original_args, **original_kw)

            return wrapped

        else:

            def partial_inner(func):
                """Case 2 - returns Case 1
                i.e., jit()(function)(*args, **kwargs).
                """

                return jit(func)

            return partial_inner

    HAS_JIT = False


def simport(modname):
    """Safely import a module without raising an error.

    Parameters
    ----------
    modname : str
        Module name needed to import.

    Returns
    -------
    _simport : tuple
        Either (True, Module) or (False, None) depending
        on whether the import succeeded.

    Notes
    -----

    Wrapping this function around an iterative context or a with
    context would allow the module to be used without necessarily
    attaching it permanently in the global namespace:

        for t,mod in simport('pandas'):
            if t:
                mod.DataFrame()
            else:
                #do alternative behavior here
            del mod #or don't del, your call

    instead of:

        t, mod = simport('pandas')
        if t:
            mod.DataFrame()
        else:
            #do alternative behavior here

    The first idiom makes it work kind of a like a with statement.

    """

    try:
        _simport = True, importlib.import_module(modname)
    except (ImportError, ModuleNotFoundError):
        _simport = False, None

    return _simport


def requires(*args, **kwargs):
    """Decorator to wrap functions with extra dependencies.

    Parameters
    ----------
    args : list
        Modules names as strings to import.
    verbose : bool
        Set as ``True`` to print a warning message on import failure.

    Returns
    -------
    inner : func
        The original function if all arg in args are importable.
    passer : func
        A function that passes if ``inner`` fails.

    """

    v = kwargs.pop("verbose", True)
    wanted = copy.deepcopy(args)

    def inner(function):
        available = [simport(arg)[0] for arg in args]
        if all(available):
            return function
        else:

            def passer(*args, **kwargs):  # noqa: ARG001
                if v:
                    missing = [arg for i, arg in enumerate(wanted) if not available[i]]
                    print(f"missing dependencies: {missing}")
                    print(f"not running {function.__name__}")
                else:
                    pass

            return passer

    return inner



"""
For completeness, we need to test a shuffled dataframe
(i.e. always send unsorted data) with:
- numeric ids
- string ids
- point dataframe
- coordinates
- check two kernel functions
- numba/nonumba
"""

import geodatasets
import geopandas
import numpy as np
import pytest

from libpysal.graph._matching import _spatial_matching
from libpysal.graph.base import Graph


@pytest.fixture(scope="session")
def stores():
    stores = geopandas.read_file(geodatasets.get_path("geoda liquor_stores")).explode(
        index_parts=False
    )
    return stores


np.random.seed(85711)
simple = np.random.random(size=(5, 2))


pulp = pytest.importorskip("pulp")
pytest.importorskip("sklearn")
default_solver = pulp.listSolvers(onlyAvailable=True)
if len(default_solver) == 0:
    raise Exception("configuration of pulp has failed, no available solvers")
default_solver = getattr(pulp, default_solver[0])()


def test_correctness_k1():
    # manual solution for simple k=1 by hungarian method
    known = np.row_stack([(0, 3), (1, 4), (2, 3), (3, 0), (3, 2), (4, 1)])
    computed = _spatial_matching(simple, n_matches=1, solver=default_solver)
    np.testing.assert_array_equal(known, np.column_stack((computed[0], computed[1])))
    computed_partial = _spatial_matching(
        simple, n_matches=1, allow_partial_match=True, solver=default_solver
    )
    # manual solution by relaxing the above
    known = np.row_stack(
        [
            (0, 2, 0.5),
            (0, 3, 0.5),
            (1, 4, 1),
            (2, 0, 0.5),
            (2, 3, 0.5),
            (3, 0, 0.5),
            (3, 2, 0.5),
            (4, 1, 1),
        ]
    )
    np.testing.assert_array_equal(known, np.column_stack(computed_partial))


@pytest.mark.network
def test_stores(stores):
    computed_heads, computed_tails, computed_weights = _spatial_matching(
        stores.head(101), n_matches=3, solver=default_solver
    )
    computed_heads_p, computed_tails_p, computed_weights_p = _spatial_matching(
        stores.head(101), allow_partial_match=True, n_matches=3, solver=default_solver
    )
    assert (computed_weights == 1).all()
    assert (computed_weights_p < 1).any()
    assert (computed_weights_p >= 0.5).all()

    for heads, tails in (
        (computed_heads, computed_tails),
        (computed_heads_p, computed_tails_p),
    ):
        _, n_by_heads = np.unique(heads, return_counts=True)
        _, n_by_tails = np.unique(tails, return_counts=True)
        assert n_by_heads.max() == 4
        assert n_by_heads.min() == 3
        assert n_by_tails.max() == 4
        assert n_by_tails.min() == 3

    gf = Graph.from_arrays(computed_heads, computed_tails, computed_weights)
    gp = Graph.from_arrays(computed_heads_p, computed_tails_p, computed_weights_p)
    for g in (gf, gp):
        assert g.asymmetry().empty
        assert g.isolates.empty


def test_returns_mip():
    *computed, mip = _spatial_matching(
        simple, n_matches=4, return_mip=True, solver=default_solver
    )
    assert mip.sol_status == 1
    assert mip.objective.value() > 0
    with pytest.warns(UserWarning, match="Problem is Infeasible"):
        *computed, mip = _spatial_matching(
            simple, n_matches=6, return_mip=True, solver=default_solver
        )
    assert mip.sol_status == -1



import geopandas as gpd
import numpy as np
import pytest
import shapely
from geopandas.testing import assert_geoseries_equal
from packaging.version import Version

from ..voronoi import voronoi, voronoi_frames


class TestVoronoi:
    def setup_method(self):
        self.points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
        self.points2 = [(10, 5), (4, 2), (5, 5)]

        self.vertices = [
            [4.21783295711061, 4.084085778781038],
            [7.519560251284979, 3.518075385494004],
            [9.464219298524961, 19.399457604620512],
            [14.982106844470032, -10.63503022227075],
            [-9.226913414477298, -4.58994413837245],
            [14.982106844470032, -10.63503022227075],
            [1.7849180090475505, 19.898032941190912],
            [9.464219298524961, 19.399457604620512],
            [1.7849180090475505, 19.898032941190912],
            [-9.226913414477298, -4.58994413837245],
        ]

        p1 = shapely.Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
        p2 = shapely.Polygon([(0, 1), (0, 2), (1, 2), (1, 1)])
        p3 = shapely.Polygon([(1, 1), (1, 2), (2, 2), (2, 1)])
        p4 = shapely.Polygon([(1, 0), (1, 1), (2, 1), (2, 0)])
        self.polygons = gpd.GeoSeries([p1, p2, p3, p4], crs="EPSG:3857")

        self.lines = gpd.GeoSeries(
            [
                shapely.LineString([(0, 0), (0, 1)]),
                shapely.LineString([(1, 1), (1, 0)]),
            ],
            crs="EPSG:3857",
        )

    def test_voronoi(self):
        with pytest.warns(
            FutureWarning, match="The 'voronoi' function is considered private"
        ):
            regions, vertices = voronoi(self.points)
        assert regions == [[1, 3, 2], [4, 5, 1, 0], [0, 1, 7, 6], [9, 0, 8]]

        np.testing.assert_array_almost_equal(vertices, self.vertices)

    def test_from_array(self):
        geoms = voronoi_frames(self.points2, as_gdf=False, return_input=False)
        expected = gpd.GeoSeries.from_wkt(
            [
                "POLYGON ((7.5 2.5, 7.5 5, 10 5, 10 2, 7.75 2, 7.5 2.5))",
                "POLYGON ((7.5 2.5, 7.75 2, 4 2, 4 3.66666666, 7.5 2.5))",
                "POLYGON ((7.5 2.5, 4 3.66666666, 4 5, 7.5 5, 7.5 2.5))",
            ],
        )
        assert_geoseries_equal(
            shapely.normalize(geoms),
            shapely.normalize(expected),
            check_less_precise=True,
        )

    def test_from_polygons(self):
        geoms = voronoi_frames(
            self.polygons, as_gdf=False, return_input=False, shrink=0.1
        )
        expected = gpd.GeoSeries.from_wkt(
            [
                "POLYGON ((0.5 1, 1 1, 1 0.5, 1 0.1, 0.1 0.1, 0.1 1, 0.5 1))",
                "POLYGON ((1 1.5, 1 1, 0.5 1, 0.1 1, 0.1 1.9, 1 1.9, 1 1.5))",
                "POLYGON ((1 1.5, 1 1.9, 1.9 1.9, 1.9 1, 1.5 1, 1 1, 1 1.5))",
                "POLYGON ((1 0.5, 1 1, 1.5 1, 1.9 1, 1.9 0.1, 1 0.1, 1 0.5))",
            ],
            crs="EPSG:3857",
        )
        assert_geoseries_equal(
            shapely.normalize(geoms),
            shapely.normalize(expected),
            check_less_precise=True,
        )

    @pytest.mark.skipif(
        Version(gpd.__version__) < Version("0.13.0"),
        reason="requires geopandas>=0.13.0",
    )
    def test_from_lines(self):
        geoms = voronoi_frames(
            self.lines, as_gdf=False, return_input=False, segment=0.1
        )
        expected = gpd.GeoSeries.from_wkt(
            [
                "POLYGON ((0.5 0.95, 0.5 0, 0 0, 0 1, 0.5 0.95))",
                "POLYGON ((0.5 0.05, 0.5 1, 1 1, 1 0, 0.5 0.05))",
            ],
            crs="EPSG:3857",
        )
        assert_geoseries_equal(geoms.simplify(0.1), expected, check_less_precise=True)

    @pytest.mark.skipif(
        Version(gpd.__version__) >= Version("0.13.0"),
        reason="requires geopandas<0.13.0",
    )
    def test_from_lines_import_error(self):
        with pytest.raises(
            ImportError,
            match="Voronoi tessellation of lines requires geopandas 0.13.0 or later.",
        ):
            voronoi_frames(self.lines, as_gdf=False, return_input=False, segment=0.1)

    def test_clip_none(self):
        geoms = voronoi_frames(
            self.points2, as_gdf=False, return_input=False, clip=None
        )
        expected = gpd.GeoSeries.from_wkt(
            [
                "POLYGON ((16 11, 16 -4, 10.75 -4, 7.5 2.5, 7.5 11, 16 11))",
                "POLYGON ((-2 -4, -2 5.6666666, 7.5 2.5, 10.75 -4, -2 -4))",
                "POLYGON ((-2 11, 7.5 11, 7.5 2.5, -2 5.66666666, -2 11))",
            ],
        )
        assert_geoseries_equal(
            shapely.normalize(geoms),
            shapely.normalize(expected),
            check_less_precise=True,
        )

    def test_clip_chull(self):
        geoms = voronoi_frames(
            self.points2, as_gdf=False, return_input=False, clip="convex_hull"
        )
        expected = gpd.GeoSeries.from_wkt(
            [
                "POLYGON ((7.5 5, 10 5, 7.5 3.75, 7.5 5))",
                "POLYGON ((6 3, 4 2, 4.5 3.5, 6 3))",
                "POLYGON ((7.5 3.75, 6 3, 4.5 3.5, 5 5, 7.5 5, 7.5 3.75))",
            ],
        )
        assert_geoseries_equal(
            shapely.normalize(geoms),
            shapely.normalize(expected),
            check_less_precise=True,
        )

    def test_clip_ahull(self):
        geoms = voronoi_frames(
            self.points2, as_gdf=False, return_input=False, clip="alpha_shape"
        )
        expected = gpd.GeoSeries.from_wkt(
            [
                "POLYGON ((7.5 5, 10 5, 7.5 3.75, 7.5 5))",
                "POLYGON ((6 3, 4 2, 4.5 3.5, 6 3))",
                "POLYGON ((7.5 3.75, 6 3, 4.5 3.5, 5 5, 7.5 5, 7.5 3.75))",
            ],
        )
        assert_geoseries_equal(
            shapely.normalize(geoms),
            shapely.normalize(expected),
            check_less_precise=True,
        )

    def test_clip_polygon(self):
        geoms = voronoi_frames(
            self.points2,
            as_gdf=False,
            return_input=False,
            clip=shapely.box(-10, -10, 10, 10),
        )
        expected = gpd.GeoSeries.from_wkt(
            [
                "POLYGON ((7.5 2.5, 7.5 10, 10 10, 10 -2.5, 7.5 2.5))",
                "POLYGON ("
                "(-10 8.333333, 7.5 2.5, 10 -2.5, 10 -10, -10 -10, -10 8.333333))",
                "POLYGON ((7.5 2.5, -10 8.33333333, -10 10, 7.5 10, 7.5 2.5))",
            ],
        )
        assert_geoseries_equal(
            shapely.normalize(geoms),
            shapely.normalize(expected),
            check_less_precise=True,
        )

    def test_clip_error(self):
        with pytest.raises(ValueError, match="Clip type 'invalid' not understood."):
            voronoi_frames(self.points2, clip="invalid")

    def test_as_gdf(self):
        geoms, polys = voronoi_frames(self.polygons, as_gdf=True, return_input=True)
        assert isinstance(geoms, gpd.GeoDataFrame)
        assert isinstance(polys, gpd.GeoDataFrame)

        with pytest.warns(
            FutureWarning,
            match="The 'as_gdf' parameter currently defaults to True but will",
        ):
            voronoi_frames(self.points2, return_input=True)

    def test_return_input(self):
        geoms, polys = voronoi_frames(self.polygons, return_input=True, as_gdf=False)
        assert isinstance(geoms, gpd.GeoSeries)
        assert polys is self.polygons

        with pytest.warns(
            FutureWarning,
            match="The 'return_input' parameter currently defaults to True but will",
        ):
            voronoi_frames(self.points2, as_gdf=True)

    def test_radius(self):
        with pytest.warns(FutureWarning, match="The 'radius' parameter is deprecated"):
            voronoi_frames(self.points2, radius=1)

    @pytest.mark.parametrize("clip", ["none", "bounds", "chull", "ahull"])
    def test_deprecated_clip(self, clip):
        with pytest.warns(
            FutureWarning,
            match=f"The '{clip}' option for the 'clip' parameter is deprecated",
        ):
            voronoi_frames(self.points2, clip=clip)



import pandas as pd


def _read_gwt(path):
    """
    Read GWT weights to Graph object

    Parameters
    ----------
    path : str
        path to GWT file

    Returns
    -------
    tuple
        focal, neighbor, weight arrays
    """
    adjacency = pd.read_csv(path, sep=r"\s+", skiprows=1, header=None)
    return adjacency[0].values, adjacency[1].values, adjacency[2].values


def _to_gwt(graph_obj, path):
    """
    Write GWT weights to Graph object

    Parameters
    ----------
    graph_obj : Graph
        Graph object
    path : str
        path to GAL file
    """
    adj = graph_obj._adjacency.reset_index()
    adj["focal"] = adj["focal"].astype(str).str.replace(" ", "_")
    adj["neighbor"] = adj["neighbor"].astype(str).str.replace(" ", "_")
    with open(path, "w") as file:
        file.write(f"0 {graph_obj.n} Unknown Unknown\n")
    adj.to_csv(path, sep=" ", header=False, index=False, mode="a", float_format="%.7f")



import numpy as np
import pytest

from ..util import lat2W
from ..weights import W

networkx = pytest.importorskip("networkx")


class TestNetworkXConverter:
    def setup_method(self):
        self.known_nx = networkx.random_regular_graph(4, 10, seed=8879)
        self.known_amat = networkx.to_numpy_array(self.known_nx)
        self.known_W = lat2W(5, 5)

    def test_round_trip(self):
        w_ = W.from_networkx(self.known_nx)
        np.testing.assert_allclose(w_.sparse.toarray(), self.known_amat)
        nx2 = w_.to_networkx()
        np.testing.assert_allclose(networkx.to_numpy_array(nx2), self.known_amat)
        nxsquare = self.known_W.to_networkx()
        np.testing.assert_allclose(
            self.known_W.sparse.toarray(), networkx.to_numpy_array(nxsquare)
        )
        w_square = W.from_networkx(nxsquare)
        np.testing.assert_allclose(
            self.known_W.sparse.toarray(), w_square.sparse.toarray()
        )



import numpy as np
import pandas as pd
import pytest
from packaging.version import Version
from scipy import __version__ as scipy_version

from libpysal import graph
from libpysal.weights.raster import testDataArray as dummy_array  # noqa: N813


class TestRaster:
    def setup_method(self):
        pytest.importorskip("xarray")
        self.da1 = dummy_array()
        self.da2 = dummy_array((1, 4, 4), missing_vals=False)
        self.da3 = self.da2.rename({"band": "layer", "x": "longitude", "y": "latitude"})
        self.data1 = pd.Series(np.ones(5))
        self.da4 = dummy_array((1, 1), missing_vals=False)
        self.da4.data = np.array([["test"]])

    @pytest.mark.skipif(
        Version(scipy_version) < Version("1.12.0"),
        reason="sparse matrix power requires scipy>=1.12.0",
    )
    def test_queen(self):
        g1 = graph.Graph.build_raster_contiguity(self.da1, rook=False, k=2, n_jobs=-1)
        assert g1[(1, -30.0, -180.0)].to_dict() == {
            (1, -90.0, 60.0): 1,
            (1, -90.0, -60.0): 1,
        }
        assert g1[(1, -30.0, 180.0)].to_dict() == {
            (1, -90.0, -60.0): 1,
            (1, -90.0, 60.0): 1,
        }
        assert g1.n == 5
        assert g1._xarray_index_names == self.da1.to_series().index.names
        assert (1, 90.0, 180.0) in g1.isolates

    def test_rook(self):
        g2 = graph.Graph.build_raster_contiguity(self.da2, rook=True)
        assert g2.neighbors[(1, -90.0, 180.0)] == (
            (1, -30.0, 180.0),
            (1, -90.0, 60.0),
        )
        assert g2.neighbors[(1, -90.0, 60.0)] == (
            (1, -30.0, 60.0),
            (1, -90.0, -60.0),
            (1, -90.0, 180.0),
        )
        assert g2.n == 16
        assert g2._xarray_index_names == self.da2.to_series().index.names

    def test_labels(self):
        coords_labels = {
            "z_label": "layer",
            "y_label": "latitude",
            "x_label": "longitude",
        }
        g3 = graph.Graph.build_raster_contiguity(
            self.da3, z_value=1, coords_labels=coords_labels
        )
        assert g3.neighbors[(1, -90.0, 180.0)] == (
            (1, -30.0, 60.0),
            (1, -30.0, 180.0),
            (1, -90.0, 60.0),
        )
        assert g3.n == 16
        assert g3._xarray_index_names == self.da3.to_series().index.names

    def test_generate_da(self):
        xarray = pytest.importorskip("xarray")
        g2 = graph.Graph.build_raster_contiguity(self.da2, rook=True, n_jobs=-1)
        da2 = g2.generate_da(self.da2.data.flatten())
        # the order may be different but shall be align-able
        a, b = xarray.align(da2, self.da2)
        xarray.testing.assert_equal(a, b)
